<?php

/**
 * @file
 * Contains \Drupal\cdek_api\Cdek.
 */

namespace Drupal\cdek_api;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Component\Utility\Unicode;
use Drupal\Component\Utility\Crypt;

/**
 * Wrapper class for CDEK API.
 *
 * See CDEK API documentation for more info.
 */
class Cdek {

  /**
   * Prefix for API data cache.
   */
  const CACHE_PREFIX = 'cdek_api_';

  /**
   * Hash for empty params set.
   */
  const EMPTY_PARAMS_HASH = 'all';

  /**
   * API auth login.
   *
   * @var string
   */
  protected $authLogin;

  /**
   * API secure password.
   *
   * @var string
   */
  protected $securePassword;

  /**
   * API data cache expire.
   *
   * @var int
   */
  protected $cacheExpire;

  /**
   * Default sender city ID.
   *
   * @var int
   */
  protected $defaultSenderCityId;

  /**
   * Default sender city postcode.
   *
   * @var int
   */
  protected $defaultSenderCityPostcode;

  /**
   * Default tariff ID.
   *
   * @var int
   */
  protected $defaultTariffId;

  /**
   * The cache object associated with the 'default' bin.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * CDEK API requests.
   *
   * @var \Drupal\cdek_api\CdekRequest
   */
  protected $cdekRequest;

  /**
   * Cdek constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\cdek_api\CdekRequest $cdek_request
   *   CDEK API requests.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache object associated with the 'default' bin.
   */
  public function __construct(ConfigFactoryInterface $config_factory, CdekRequest $cdek_request, CacheBackendInterface $cache) {
    $this->cdekRequest = $cdek_request;
    $this->cache = $cache;
    // Get the configuration values.
    $config = $config_factory->get('cdek_api.settings');
    $this->authLogin = $config->get('auth_login');
    $this->securePassword = $config->get('secure_password');
    $this->cacheExpire = $config->get('cache_expire');
    $this->defaultSenderCityId = $config->get('sender_city_id');
    $this->defaultSenderCityPostcode = $config->get('sender_city_postcode');
    $this->defaultTariffId = $config->get('tariff_id');
  }

  /**
   * Load order pickup points.
   *
   * @param array $params
   *   API function params for filtered results.
   *   The following parameters are allowed:
   *   - cityid - City code in CDEK system.
   *   - citypostcode - City postcode string, cityid is more priority.
   *   - type - Pickup point type. Allowed values: "PVZ", "POSTOMAT", "ALL".
   *     Default value is "PVZ".
   *   - countryid - Country code in CDEK system.
   *   - regionid - Region code in CDEK system.
   *   - havecashless - (boolean) Is there a payment terminal.
   *   - isdressingroom - (boolean) Is there a fitting room in the pickup point.
   *   - weightmax - The maximum weight that can be sent to the pickup point.
   *   - allowedcod - (boolean) CASH is allowed in the pickup point.
   *
   * @return array|null
   *   Pickup points array (each point is a stdClass object) or NULL on failure.
   */
  public function getOrderPickupPoints(array $params = []) {
    static $pickup_points = [];
    $hash = $this->generateParamsHash($params);

    // Check static data.
    if (isset($pickup_points[$hash])) {
      return $pickup_points[$hash];
    }
    // Check cache data.
    $cid = "order_pickup_points:$hash";
    if ($this->isCached() && ($cache_data = $this->getCacheData($cid)) !== FALSE) {
      return $pickup_points[$hash] = $cache_data;
    }

    // Sending API request.
    $result = $this->cdekRequest->sendOrderPickupPointsRequest($params);
    if ($result === NULL) {
      return NULL;
    }
    // Parse response XML with SimpleXML extension and check for errors.
    $result = $this->parseOrderPickupPointsXml($result);
    if ($result === NULL) {
      return NULL;
    }

    // Set persistent and static cache.
    $pickup_points[$hash] = $result;
    if ($this->isCached()) {
      $this->setCacheData($cid, $pickup_points[$hash]);
    }
    return $pickup_points[$hash];
  }

  /**
   * Load countries.
   *
   * @return array|null
   *   Country names array keyed by country code, or NULL on failure.
   */
  public function getCountries() {
    static $countries;

    // Check static data.
    if (isset($countries)) {
      return $countries;
    }
    // Check cache data.
    if ($this->isCached() && ($cache_data = $this->getCacheData('countries')) !== FALSE) {
      return $countries = $cache_data;
    }

    // Load order pickup points.
    $pickup_points = $this->getOrderPickupPoints();
    if ($pickup_points === NULL) {
      return NULL;
    }

    // Get countries from pickup points.
    $countries = [];
    foreach ($pickup_points as $pickup_point) {
      if (!isset($countries[$pickup_point->countryCode])) {
        $countries[$pickup_point->countryCode] = $pickup_point->countryName;
      }
    }
    // Sort countries in alphabet.
    asort($countries);
    // Set persistent cache.
    if ($this->isCached()) {
      $this->setCacheData('countries', $countries);
    }
    return $countries;
  }

  /**
   * Load regions by country.
   *
   * @param int $country
   *   Country code in CDEK system.
   *
   * @return array|null
   *   Region names array keyed by region code, or NULL on failure.
   */
  public function getRegions($country = NULL) {
    static $regions = [];

    // Prepare API request params.
    $params = ['countryid' => $country];
    if ($country === NULL) {
      $country = self::EMPTY_PARAMS_HASH;
      unset($params['countryid']);
    }

    // Check static data.
    if (isset($regions[$country])) {
      return $regions[$country];
    }
    // Check cache data.
    $cid = "regions:$country";
    if ($this->isCached() && ($cache_data = $this->getCacheData($cid)) !== FALSE) {
      return $regions[$country] = $cache_data;
    }

    // Load order pickup points.
    $pickup_points = $this->getOrderPickupPoints($params);
    if ($pickup_points === NULL) {
      return NULL;
    }

    // Get regions from pickup points.
    $regions[$country] = [];
    foreach ($pickup_points as $pickup_point) {
      if (!isset($regions[$country][$pickup_point->regionCode])) {
        $regions[$country][$pickup_point->regionCode] = $pickup_point->regionName;
      }
    }
    // Sort regions in alphabet.
    asort($regions[$country]);
    // Set persistent cache.
    if ($this->isCached()) {
      $this->setCacheData($cid, $regions[$country]);
    }
    return $regions[$country];
  }

  /**
   * Load cities by country and/or region.
   *
   * @param int $country
   *   Country code in CDEK system.
   * @param int $region
   *   Region code in CDEK system.
   *
   * @return array|null
   *   City names array keyed by city code, or NULL on failure.
   */
  public function getCities($country = NULL, $region = NULL) {
    static $cities = [];

    // Prepare API request params.
    $params = ['countryid' => $country, 'regionid' => $region];
    if ($country === NULL) {
      $country = self::EMPTY_PARAMS_HASH;
      unset($params['countryid']);
    }
    if ($region === NULL) {
      $region = self::EMPTY_PARAMS_HASH;
      unset($params['regionid']);
    }

    // Check static data.
    if (isset($cities[$country][$region])) {
      return $cities[$country][$region];
    }
    // Check cache data.
    $cid = "cities:$country:$region";
    if ($this->isCached() && ($cache_data = $this->getCacheData($cid)) !== FALSE) {
      return $cities[$country][$region] = $cache_data;
    }

    // Load order pickup points.
    $pickup_points = $this->getOrderPickupPoints($params);
    if ($pickup_points === NULL) {
      return NULL;
    }

    // Get cities from pickup points.
    $cities[$country][$region] = [];
    foreach ($pickup_points as $pickup_point) {
      if (!isset($cities[$country][$region][$pickup_point->cityCode])) {
        $cities[$country][$region][$pickup_point->cityCode] = $pickup_point->city;
      }
    }
    // Sort cities in alphabet.
    asort($cities[$country][$region]);
    // Set persistent cache.
    if ($this->isCached()) {
      $this->setCacheData($cid, $cities[$country][$region]);
    }
    return $cities[$country][$region];
  }

  /**
   * Calculate delivery rate by params.
   *
   * @param array $params
   *   API function params to calculate.
   *   The following parameters are allowed:
   *   - authLogin - Login for CDEK account. Set TRUE to use the login
   *     specified in the configuration.
   *   - secure - Password for CDEK account. If authLogin is set to TRUE,
   *     then the password will be taken from the configuration.
   *   - dateExecute - Planning date of shipment in the format YYYY-ММ-DD.
   *   - senderCityId - Origin city code in CDEK system.
   *   - senderCityPostCode - Zip code of the city of origin.
   *   - receiverCityId - Destination city code in CDEK system.
   *   - receiverCityPostCode - Zip code of the destination city.
   *   - tariffId - Code of chosen tariff.
   *   - modeId - Chosen delivery mode.
   *   - goods - Array with elements each of which contains an
   *     associative array with keys:
   *       - weight - float, piece weight, kg.
   *       - length - integer, piece length, sm.
   *       - width - integer, piece width, sm.
   *       - height - integer, depth, sm.
   *     Or with such kind of parameters:
   *       - weight - float, piece weight, kg.
   *       - volume - float, piece volume, cum.
   *
   * @return object|null
   *   Delivery rate information object (stdClass), or NULL on failure.
   *
   * @throws \Exception
   *   CDEK API exception.
   */
  public function calculateDeliveryRate($params = []) {
    // TODO: Define a constant for "version" or move it to a configuration.
    // TODO: Add "tariffList" and "services" params support.
    // Merge to defaults.
    $params += [
      'version' => '1.0',
      'senderCityId' => $this->defaultSenderCityId,
      'senderCityPostCode' => $this->defaultSenderCityPostcode,
      'tariffId' => $this->defaultTariffId,
    ];

    // Use authorization if necessary.
    if (isset($params['authLogin']) && $params['authLogin'] === TRUE) {
      $params['authLogin'] = $this->authLogin;
      $params['secure'] = $this->securePassword;
    }
    // Process "secure" param.
    if (isset($params['secure'])) {
      if (!isset($params['dateExecute'])) {
        $params['dateExecute'] = date('Y-m-d', time());
      }
      $params['secure'] = md5($params['dateExecute'] . '&' . $params['secure']);
    }
    // Filter empty values.
    $params = array_filter($params);

    // Sending API request.
    $result = $this->cdekRequest->sendDeliveryRateCalculationRequest($params);
    if ($result === NULL) {
      return NULL;
    }
    // TODO: Define rate info class.
    // TODO: Add serializer for rate info.
    $data = json_decode($result);
    // Check API errors.
    if (isset($data->error)) {
      // Throw exception by first error.
      // TODO: Custom exception class.
      $error = array_shift($data->error);
      throw new \Exception($error->text, $error->code);
    }
    return $data->result;
  }

  /**
   * Verifying that the data is cached.
   *
   * @return bool
   *   API data is cached (TRUE) or not (FALSE).
   */
  public function isCached() {
    return !empty($this->cacheExpire);
  }

  /**
   * Stores API data in the persistent cache.
   *
   * @param string $cid
   *   The cache ID of the data to store.
   * @param mixed $data
   *   The data to store in the cache.
   */
  protected function setCacheData($cid, $data) {
    $cid = self::CACHE_PREFIX . $cid;
    $this->cache->set($cid, $data, time() + $this->cacheExpire);
  }

  /**
   * Returns API data from the persistent cache.
   *
   * @param string $cid
   *   The cache ID of the data to retrieve.
   *
   * @return object|false
   *   The cache data or FALSE on failure.
   */
  protected function getCacheData($cid) {
    $cid = self::CACHE_PREFIX . $cid;
    $cache = $this->cache->get($cid);
    return $cache && isset($cache->data) ? $cache->data : FALSE;
  }

  /**
   * Calculates hash for params set.
   *
   * @param array $params
   *   A set of params to calculate a hash.
   *
   * @return string
   *   Params set hash.
   */
  protected function generateParamsHash(array $params) {
    if (empty($params)) {
      return self::EMPTY_PARAMS_HASH;
    }
    ksort($params);
    $params = serialize($params);
    return Crypt::hashBase64($params);
  }

  /**
   * Parse order pickup points from XML.
   *
   * @param string $xml
   *   XML containing order pickup points.
   *
   * @return array|null
   *   Associative array of order pickup point objects keyed by code or NULL
   */
  protected function parseOrderPickupPointsXml($xml) {
    $data = simplexml_load_string($xml);
    if ($data === FALSE) {
      return NULL;
    }

    // TODO: Define pickup point class.
    // TODO: Add serializer for pickup points.
    // TODO: Process pickup point children.
    $pickup_points = [];
    /** @var \SimpleXMLElement $child */
    foreach ($data->children() as $child) {
      // Create class from the values of the attributes.
      $pickup_point = new \stdClass();
      foreach ($child->attributes() as $name => $value) {
        $name = Unicode::lcfirst($name);
        $value = trim((string) $value);
        $pickup_point->$name = $value === '' ? NULL : $value;
      }
      // Add pickup point object in result array.
      if (isset($pickup_point->code)) {
        $pickup_points[$pickup_point->code] = $pickup_point;
      }
    }
    return $pickup_points;
  }

}
