<?php

/**
 * @file
 * Contains \Drupal\cdek_api\CdekRequest.
 */

namespace Drupal\cdek_api;

use Drupal\Core\Config\ConfigFactoryInterface;
use Psr\Http\Message\ResponseInterface;
use GuzzleHttp\Client;

/**
 * CDEK API requests.
 */
class CdekRequest {

  /**
   * Request path for order pickup points list.
   */
  const ORDER_PICKUP_POINTS_LIST_PATH = 'pvzlist.php';

  /**
   * Request path for delivery rate calculation.
   */
  const DELIVERY_RATE_CALCULATION_PATH = 'calculator/calculate_price_by_json.php';

  /**
   * Integration API base url.
   *
   * @var string
   */
  protected $integrationApiBasePath;

  /**
   * API base url.
   *
   * @var string
   */
  protected $apiBasePath;

  /**
   * API data request timeout.
   *
   * @var float
   */
  protected $requestTimeout;

  /**
   * CdekRequest constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \GuzzleHttp\Client $http_client
   *   A guzzle http client instance.
   */
  public function __construct(ConfigFactoryInterface $config_factory, Client $http_client) {
    $this->httpClient = $http_client;
    // Get the configuration values.
    $config = $config_factory->get('cdek_api.settings');
    $this->integrationApiBasePath = $config->get('integration_base_path');
    $this->apiBasePath = $config->get('base_path');
    $this->requestTimeout = $config->get('request_timeout');
  }

  /**
   * Returns response from "order pickup points list" API request.
   *
   * @param array $params
   *   See \Drupal\cdek_api\Cdek::getOrderPickupPoints() for details.
   *
   * @return string|null
   *   Response body or NULL on failure.
   */
  public function sendOrderPickupPointsRequest(array $params = []) {
    // Request options and URI.
    $uri = $this->integrationApiBasePath . '/' . self::ORDER_PICKUP_POINTS_LIST_PATH;
    $options['timeout'] = (float) $this->requestTimeout;
    $options['query'] = $params;
    // Sending GET request.
    $response = $this->httpClient->get($uri, $options);
    // Check the success of the API request.
    return $this->isSuccessfulRequest($response) ? (string) $response->getBody() : NULL;
  }

  /**
   * Returns response from "calculate delivery rate" API request.
   *
   * @param array $params
   *   See \Drupal\cdek_api\Cdek::calculateDeliveryRate() for details.
   *
   * @return string|null
   *   Response body or NULL on failure.
   */
  public function sendDeliveryRateCalculationRequest(array $params = []) {
    // Request options and URI.
    $uri = $this->apiBasePath . '/' . self::DELIVERY_RATE_CALCULATION_PATH;
    $options['headers'] = ['Content-Type' => 'application/json'];
    $options['timeout'] = (float) $this->requestTimeout;
    $options['body'] = json_encode($params);
    // Sending POST request.
    $response = $this->httpClient->post($uri, $options);
    // Check the success of the API request.
    return $this->isSuccessfulRequest($response) ? (string) $response->getBody() : NULL;
  }

  /**
   * Check the success of the API request.
   *
   * @param \Psr\Http\Message\ResponseInterface $response
   *   Response of the API request.
   *
   * @return bool
   *   The request is successful (TRUE) or not (FALSE).
   */
  protected function isSuccessfulRequest(ResponseInterface $response) {
    return $response->getStatusCode() == '200' && $response->getReasonPhrase() == 'OK';
  }

}
